"""Exports a YOLOv5 *.pt model to ONNX and TorchScript formats

Usage:
   # Invke from top dir
   $ python models/prune_TP.py --model weights/yolov5s.pt 
"""

import argparse
import sys
import time
import numpy as np
sys.path.append('./')  # to run '$ python *.py' files in subdirectories

import torch
import torch.nn as nn
import torch_pruning as tp

import models
from models.experimental import attempt_load
from utils.activations import Hardswish, SiLU
from utils.general import set_logging, check_img_size
import os

SPARSITY = 0.5
def prune_model(model, img):
    model.cpu()
    model.eval()
    img.cpu()
    DG = tp.DependencyGraph().build_dependency( model, img)
    print("------------ BEGIN : DG._module_to_name ----------------------")
    print(DG._module_to_name)
    print("------------ END   : DG._module_to_name ----------------------")
    print("------------ BEGIN : DG.module_to_node  ---------------------")
    print(DG.module_to_node)
    print("------------ END   : DG.module_to_node  ---------------------")
    def prune_conv(conv, pruned_prob):
        weight = conv.weight.detach().cpu().numpy()
        out_channels = weight.shape[0]
        L1_norm = np.sum( np.abs(weight), axis=(1,2,3))
        num_pruned = int(out_channels * pruned_prob)
        prune_index = np.argsort(L1_norm)[:num_pruned].tolist() # remove filters with small L1-Norm
        print("======== prune_conv() ========")
        print("prune_index = ", prune_index)
        print("num_pruned = ", num_pruned)
        try:
            plan = DG.get_pruning_plan(conv, tp.prune_conv, prune_index)
        except Exception as e:
            raise e
            #print("ERROR : ", e)
            #return
        else:
            plan.exec()

    print("***** BEGIN PRUNING ***********************************************************")
    for m in model.modules():
        for name, mm in m.named_children():
            print("LEVEL 1 : ", name, mm)
            for name2, mm2 in mm.named_children():
                print("____ LEVEL 2 : ", name2, mm2)
                for name3, mm3 in mm2.named_children():
                    print("________ LEVEL 3 : ", name3,mm3)
                    if isinstance(mm3, nn.Conv2d) and name3 != 'conv3':
                        print("Conv2d detected. Attempting to prune......")
                        prune_conv(mm3, SPARSITY)
                    print("________ END CHILD LEVEL 3 ______________________")
                print("____ END CHILD LEVEL 2 ______________________")
            print("END CHILD LEVEL 1 ______________________")
    print("***** END PRUNING ***********************************************************")

    print("BEGIN: Torch jit of pruned model ************ ")
    model_output = torch.jit.trace(model, torch_image)
    print("Torch jit of pruned model succeeded *********** ")
    return model

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--weights', type=str, default='./yolov5s.pt', help='weights path')  # from yolov5/models/
    parser.add_argument('--img-size', nargs='+', type=int, default=[640, 640], help='image size')  # height, width
    parser.add_argument('--batch-size', type=int, default=1, help='batch size')
    opt = parser.parse_args()
    opt.img_size *= 2 if len(opt.img_size) == 1 else 1  # expand
    print(opt)
    set_logging()
    t = time.time()

    # Load PyTorch model
    model = attempt_load(opt.weights, map_location=torch.device('cpu'))  # load FP32 model

    # Checks
    gs = int(max(model.stride))  # grid size (max stride)
    opt.img_size = [check_img_size(x, gs) for x in opt.img_size]  # verify img_size are gs-multiples

    # Input
    img = torch.zeros(opt.batch_size, 3, *opt.img_size)  # image size(1,3,320,192) iDetection
    # Update model
    for k, m in model.named_modules():
        m._non_persistent_buffers_set = set()  # pytorch 1.6.0 compatibility
        try:
            isinstance(m.act, nn.LeakyReLU)
        except:
            pass
        else:
            if isinstance(m.act, nn.LeakyReLU):
                #m.act = Hardswish() ## Commenting for NNI Model speedup error issue
                m.act = nn.ReLU()
        if isinstance(m, models.common.Conv):  # assign export-friendly activations
            if isinstance(m.act, nn.Hardswish):
                #m.act = Hardswish() ## Commenting for NNI Model speedup error issue
                m.act = nn.ReLU()
            elif isinstance(m.act, nn.SiLU):
                #m.act = SiLU() ### may be thiss will also be a problem in NNJ
                m.act = nn.ReLU()
        # elif isinstance(m, models.yolo.Detect):
        #     m.forward = m.forward_export  # assign forward (optional)
    model.model[-1].export = True  # set Detect() layer export=True

    new_model = prune_model(model, img)

